@@ -23,6 +23,8 @@
 #include "config.h"
 #endif
 
+#define GDK_PIXBUF_ENABLE_BACKEND
+
 #include "eog-image.h"
 #include "eog-image-private.h"
 #include "eog-debug.h"
@@ -46,6 +48,7 @@
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
+#include <gdk-pixbuf/gdk-pixbuf-io.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomeui/gnome-thumbnail.h>
 
@@ -110,6 +113,13 @@ eog_image_free_mem_private (EogImage *image)
 
 		priv->exif_chunk_len = 0;
 
+#ifdef HAVE_EXEMPI
+		if (priv->xmp != NULL) {
+			xmp_free (priv->xmp);
+			priv->xmp = NULL;
+		}
+#endif
+
 #ifdef HAVE_LCMS
 		if (priv->profile != NULL) {
 			cmsCloseProfile (priv->profile);
@@ -160,6 +170,11 @@ eog_image_dispose (GObject *object)
 		priv->trans = NULL;
 	}
 
+	if (priv->trans_autorotate) {
+		g_object_unref (priv->trans_autorotate);
+		priv->trans_autorotate = NULL;
+	}
+
 	if (priv->undo_stack) {
 		g_slist_foreach (priv->undo_stack, (GFunc) g_object_unref, NULL);
 		g_slist_free (priv->undo_stack);
@@ -231,15 +246,20 @@ eog_image_init (EogImage *img)
 	img->priv->modified = FALSE;
 	img->priv->status_mutex = g_mutex_new ();
 	img->priv->status = EOG_IMAGE_STATUS_UNKNOWN;
+	img->priv->prev_status = EOG_IMAGE_STATUS_UNKNOWN;
+	img->priv->is_monitored = FALSE;
 	img->priv->undo_stack = NULL;
 	img->priv->trans = NULL;
+	img->priv->trans_autorotate = NULL;
 	img->priv->data_ref_count = 0;
 #ifdef HAVE_EXIF
 	img->priv->orientation = 0;
 	img->priv->autorotate = FALSE;
 	img->priv->exif = NULL;
 #endif
-
+#ifdef HAVE_EXEMPI
+	img->priv->xmp = NULL;
+#endif
 #ifdef HAVE_LCMS
 	img->priv->profile = NULL;
 #endif
@@ -300,7 +320,7 @@ eog_image_update_exif_data (EogImage *image)
 	bo = exif_data_get_byte_order (priv->exif);
 
 	/* Update image width */	
-	entry = exif_content_get_entry (priv->exif->ifd [EXIF_IFD_EXIF], EXIF_TAG_PIXEL_X_DIMENSION);
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_PIXEL_X_DIMENSION);
 	if (entry != NULL && (priv->width >= 0)) {
 		if (entry->format == EXIF_FORMAT_LONG)
 			exif_set_long (entry->data, bo, priv->width);
@@ -309,9 +329,9 @@ eog_image_update_exif_data (EogImage *image)
 		else
 			g_warning ("Exif entry has unsupported size");
 	}
-	
+
 	/* Update image height */	
-	entry = exif_content_get_entry (priv->exif->ifd [EXIF_IFD_EXIF], EXIF_TAG_PIXEL_Y_DIMENSION);
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_PIXEL_Y_DIMENSION);
 	if (entry != NULL && (priv->height >= 0)) {
 		if (entry->format == EXIF_FORMAT_LONG)
 			exif_set_long (entry->data, bo, priv->height);
@@ -320,6 +340,19 @@ eog_image_update_exif_data (EogImage *image)
 		else
 			g_warning ("Exif entry has unsupported size");
 	}
+	
+	/* Update image orientation */	
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_ORIENTATION);
+	if (entry != NULL) {
+		if (entry->format == EXIF_FORMAT_LONG)
+			exif_set_long (entry->data, bo, 1);
+		else if (entry->format == EXIF_FORMAT_SHORT)
+			exif_set_short (entry->data, bo, 1);
+		else
+			g_warning ("Exif entry has unsupported size");
+
+		priv->orientation = 1;
+	}
 #endif
 }
 
@@ -383,6 +416,40 @@ eog_image_real_transform (EogImage     *img,
 	}
 }
 
+static gboolean
+check_loader_threadsafety (GdkPixbufLoader *loader, gboolean *result)
+{
+	GdkPixbufFormat *format;
+	gboolean ret_val = FALSE;
+
+	format = gdk_pixbuf_loader_get_format (loader);
+	if (format) {
+		ret_val = TRUE;
+		if (result) 
+		/* FIXME: We should not be accessing this struct internals
+ 		 * directly. Keep track of bug #469209 to fix that. */
+			*result = format->flags & GDK_PIXBUF_FORMAT_THREADSAFE;
+	}
+
+	return ret_val;	
+}
+
+static void
+eog_image_pre_size_prepared (GdkPixbufLoader *loader,
+			     gint width,
+			     gint height,
+			     gpointer data)
+{
+	EogImage *img;
+
+	eog_debug (DEBUG_IMAGE_LOAD);
+
+	g_return_if_fail (EOG_IS_IMAGE (data));
+
+	img = EOG_IMAGE (data);
+	check_loader_threadsafety (loader, &img->priv->threadsafe_format);
+}
+
 static void
 eog_image_size_prepared (GdkPixbufLoader *loader, 
 			 gint             width, 
@@ -405,7 +472,9 @@ eog_image_size_prepared (GdkPixbufLoader *loader,
 	g_mutex_unlock (img->priv->status_mutex);
 
 #ifdef HAVE_EXIF
-	if (!img->priv->autorotate)
+	if (img->priv->threadsafe_format && (!img->priv->autorotate || img->priv->exif))
+#else
+	if (img->priv->threadsafe_format)
 #endif
 		g_signal_emit (img, signals[SIGNAL_SIZE_PREPARED], 0, width, height);
 }
@@ -433,20 +502,21 @@ eog_image_needs_transformation (EogImage *img)
 {
 	g_return_val_if_fail (EOG_IS_IMAGE (img), FALSE);
 
-	return (img->priv->trans != NULL);
+	return (img->priv->trans != NULL || img->priv->trans_autorotate != NULL);
 }
 
 static gboolean
 eog_image_apply_transformations (EogImage *img, GError **error)
 {
 	GdkPixbuf *transformed = NULL;
+	EogTransform *composition = NULL;
 	EogImagePrivate *priv;
 
 	g_return_val_if_fail (EOG_IS_IMAGE (img), FALSE);
 	
 	priv = img->priv;
 
-	if (priv->trans == NULL) {
+	if (priv->trans == NULL && priv->trans_autorotate == NULL) {
 		return TRUE;
 	}
 
@@ -459,28 +529,44 @@ eog_image_apply_transformations (EogImage *img, GError **error)
 		return FALSE;
 	}
 
-	g_assert (priv->image != NULL);
-
-	if (priv->trans != NULL) {
-		transformed = eog_transform_apply (priv->trans, priv->image, NULL);
+	if (priv->trans != NULL && priv->trans_autorotate != NULL) {
+		composition = eog_transform_compose (priv->trans, 
+						     priv->trans_autorotate);
+	} else if (priv->trans != NULL) {
+		composition = g_object_ref (priv->trans);
+	} else if (priv->trans_autorotate != NULL) {
+		composition = g_object_ref (priv->trans_autorotate);
 	}
 
+	if (composition != NULL) {
+		transformed = eog_transform_apply (composition, priv->image, NULL);
+	}
+
+	g_object_unref (priv->image);
+	priv->image = transformed;
+
 	if (transformed != NULL) {
-		g_object_unref (priv->image);
-		priv->image = transformed;
-
 		priv->width = gdk_pixbuf_get_width (priv->image);
 		priv->height = gdk_pixbuf_get_height (priv->image);
-	}
+	} else {
+		g_set_error (error, 
+			     EOG_IMAGE_ERROR, 
+			     EOG_IMAGE_ERROR_GENERIC,
+			     _("Transformation failed."));
+ 	}
+
+	g_object_unref (composition);
 
 	return (transformed != NULL);
 }
 
-static GnomeVFSFileSize
-eog_image_determine_file_bytes (EogImage *img, GError **error)
+static void 
+eog_image_get_file_info (EogImage *img, 
+			 GnomeVFSFileSize *bytes,
+			 gchar **mime_type, 
+			 GError **error)
 {
 	GnomeVFSFileInfo *info;
-	GnomeVFSFileSize bytes;
 	GnomeVFSResult result;
 
 	info = gnome_vfs_file_info_new ();
@@ -488,23 +574,31 @@ eog_image_determine_file_bytes (EogImage *img, GError **error)
 	result = gnome_vfs_get_file_info_uri (img->priv->uri,
 					      info,
 					      GNOME_VFS_FILE_INFO_DEFAULT |
+					      GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
 					      GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
 
 	if ((result != GNOME_VFS_OK) || 
-	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE) == 0) {
-		bytes = 0;
+	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE) == 0 ||
+	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE) == 0) {
+		if (bytes)
+			*bytes = 0;
+
+		if (mime_type)
+			*mime_type = NULL;
 
 		g_set_error (error, 
 			     EOG_IMAGE_ERROR, 
 			     EOG_IMAGE_ERROR_VFS,
 			     gnome_vfs_result_to_string (result));
 	} else {
-		bytes = info->size;
+		if (bytes)
+			*bytes = info->size;
+
+		if (mime_type)
+			*mime_type = g_strdup (info->mime_type);
 	}
 
 	gnome_vfs_file_info_unref (info);
-
-	return bytes;
 }
 
 #ifdef HAVE_LCMS
@@ -730,15 +824,7 @@ eog_image_real_autorotate (EogImage *img)
 		lookup[priv->orientation - 1] : EOG_TRANSFORM_NONE);
 
 	if (type != EOG_TRANSFORM_NONE) {
-		EogTransform *trans = eog_transform_new (type);
-
-		if (priv->trans == NULL) {
-			priv->trans = g_object_ref (trans);
-		} else {
-			eog_transform_compose (priv->trans, trans);
-		}
-
-		g_object_unref (trans);
+		img->priv->trans_autorotate = eog_transform_new (type);
 	}
 
 	/* Disable auto orientation for next loads */
@@ -755,6 +841,20 @@ eog_image_autorotate (EogImage *img)
 }
 #endif
 
+#ifdef HAVE_EXEMPI
+static void
+eog_image_set_xmp_data (EogImage *img, EogMetadataReader *md_reader)
+{
+	EogImagePrivate *priv;
+
+	g_return_if_fail (EOG_IS_IMAGE (img));
+
+	priv = img->priv;
+
+	priv->xmp = eog_metadata_reader_get_xmp_data (md_reader);
+}
+#endif
+
 static void
 eog_image_set_exif_data (EogImage *img, EogMetadataReader *md_reader)
 {
@@ -770,16 +870,19 @@ eog_image_set_exif_data (EogImage *img, EogMetadataReader *md_reader)
 	priv->exif_chunk = NULL;
 	priv->exif_chunk_len = 0;
 
-	/* EXIF data is already available, set the image
-           orientation and emit size prepared signal */
+	/* EXIF data is already available, set the image orientation */
 	if (priv->autorotate) {
 		eog_image_set_orientation (img);
 
-		g_signal_emit (img, 
-			       signals[SIGNAL_SIZE_PREPARED], 
-			       0, 
-			       priv->width, 
-			       priv->height);
+		/* Emit size prepared signal if we have the size */
+		if (priv->width > 0 &&
+		    priv->height > 0) {
+			g_signal_emit (img, 
+				       signals[SIGNAL_SIZE_PREPARED], 
+				       0, 
+				       priv->width, 
+				       priv->height);
+		}
 	}
 #else
 	eog_metadata_reader_get_exif_chunk (md_reader, 
@@ -792,7 +895,6 @@ eog_image_set_exif_data (EogImage *img, EogMetadataReader *md_reader)
  * Attempts to get the image dimensions from the thumbnail.
  * Returns FALSE if this information is not found.
  **/
-
 static gboolean
 eog_image_get_dimension_from_thumbnail (EogImage *image,
 			                gint     *width,
@@ -831,6 +933,7 @@ eog_image_real_load (EogImage *img,
 	GnomeVFSResult result;
 	EogMetadataReader *md_reader = NULL;
 	GdkPixbufFormat *format;
+	gchar *mime_type;
 	GdkPixbufLoader *loader = NULL;
 	guchar *buffer;
 	gboolean failed = FALSE;
@@ -840,8 +943,6 @@ eog_image_real_load (EogImage *img,
 	gboolean read_only_dimension = (data2read & EOG_IMAGE_DATA_DIMENSION) && 
 					!read_image_data;
 
-	g_assert (error == NULL || *error == NULL);
-
 	priv = img->priv;
 
  	g_assert (!read_image_data || priv->image == NULL);
@@ -851,9 +952,12 @@ eog_image_real_load (EogImage *img,
 		priv->file_type = NULL;
 	}
 
-	priv->bytes = eog_image_determine_file_bytes (img, error);
+	priv->threadsafe_format = FALSE;
 
-	if (priv->bytes == 0 && (error == NULL || *error != NULL)) {
+	eog_image_get_file_info (img, &priv->bytes, &mime_type, error);
+
+	if (*error) {
+		g_free (mime_type);
 		return FALSE;
 	}
 
@@ -869,6 +973,7 @@ eog_image_real_load (EogImage *img,
 			priv->width = width;
 			priv->height = height;
 
+			g_free (mime_type);
 			return TRUE;
 		}
 	}
@@ -876,6 +981,8 @@ eog_image_real_load (EogImage *img,
 	result = gnome_vfs_open_uri (&handle, priv->uri, GNOME_VFS_OPEN_READ);
 
 	if (result != GNOME_VFS_OK) {
+		g_free (mime_type);
+
 		g_set_error (error, 
 			     EOG_IMAGE_ERROR, 
 			     EOG_IMAGE_ERROR_VFS,
@@ -885,9 +992,31 @@ eog_image_real_load (EogImage *img,
 	}
 	
 	buffer = g_new0 (guchar, EOG_IMAGE_READ_BUFFER_SIZE);
-
+	
 	if (read_image_data || read_only_dimension) {
-		loader = gdk_pixbuf_loader_new ();
+		gboolean checked_threadsafety = FALSE;
+
+		loader = gdk_pixbuf_loader_new_with_mime_type (mime_type, error);
+	
+		if (*error) {
+			g_error_free (*error);
+			*error = NULL;
+
+			loader = gdk_pixbuf_loader_new (); 
+		} else {
+			/* The mimetype-based loader should know the
+			 * format here already. */
+			checked_threadsafety = check_loader_threadsafety (loader, &priv->threadsafe_format);
+		}
+
+		/* This is used to detect non-threadsafe loaders and disable
+ 		 * any possible asyncronous task that could bring deadlocks
+ 		 * to image loading process. */
+		if (!checked_threadsafety)
+			g_signal_connect (loader, 
+					  "size-prepared",
+					  G_CALLBACK (eog_image_pre_size_prepared),
+					  img);
 
 		g_signal_connect_object (G_OBJECT (loader), 
 					 "size-prepared", 
@@ -895,6 +1024,7 @@ eog_image_real_load (EogImage *img,
 					 img, 
 					 0);
         }
+	g_free (mime_type);
 
 	while (!priv->cancel_loading) {
 		result = gnome_vfs_read (handle, 
@@ -940,11 +1070,12 @@ eog_image_real_load (EogImage *img,
 					break;
 				}
 
-				g_signal_emit (img, 
-					       signals[SIGNAL_SIZE_PREPARED], 
-					       0, 
-					       priv->width, 
-					       priv->height);
+				if (priv->threadsafe_format)
+					g_signal_emit (img, 
+						       signals[SIGNAL_SIZE_PREPARED], 
+						       0, 
+						       priv->width, 
+						       priv->height);
                         }
 
 			first_run = FALSE;
@@ -956,9 +1087,14 @@ eog_image_real_load (EogImage *img,
 			if (eog_metadata_reader_finished (md_reader)) {
 				if (set_metadata) {
 					eog_image_set_exif_data (img, md_reader);
+
 #ifdef HAVE_LCMS
 					eog_image_set_icc_data (img, md_reader);
 #endif
+
+#ifdef HAVE_EXEMPI
+					eog_image_set_xmp_data (img, md_reader);
+#endif
 					set_metadata = FALSE;
 				}
 
@@ -977,8 +1113,12 @@ eog_image_real_load (EogImage *img,
 		if (failed) {
 			gdk_pixbuf_loader_close (loader, NULL);
 		} else if (!gdk_pixbuf_loader_close (loader, error)) {
-			failed = TRUE;
-		}
+			if (gdk_pixbuf_loader_get_pixbuf (loader) != NULL) {
+				/* Clear error in order to support partial
+				 * images as well. */
+				g_clear_error (error);
+			}
+	        }
         }
 
 	g_free (buffer);
@@ -988,8 +1128,7 @@ eog_image_real_load (EogImage *img,
 	failed = (failed ||
 		  priv->cancel_loading || 
 		  bytes_read_total == 0 || 
-		  *error != NULL || 
-		  priv->cancel_loading);
+		  *error != NULL);
 
 	if (failed) {
 		if (priv->cancel_loading) {
@@ -1018,6 +1157,15 @@ eog_image_real_load (EogImage *img,
 			if (format != NULL) {
 				priv->file_type = gdk_pixbuf_format_get_name (format);
 			}
+
+			/* If it's non-threadsafe loader, then trigger window 
+ 			 * showing in the end of the process. */
+			if (!priv->threadsafe_format)
+				g_signal_emit (img, 
+					       signals[SIGNAL_SIZE_PREPARED], 
+					       0, 
+					       priv->width, 
+					       priv->height);
 		}
 	}
 
@@ -1030,6 +1178,14 @@ eog_image_real_load (EogImage *img,
 		md_reader = NULL;
 	}	
 
+	/* Catch-all in case of poor-error reporting */
+	if (failed && *error == NULL) {
+		g_set_error (error, 
+			     EOG_IMAGE_ERROR,
+			     EOG_IMAGE_ERROR_GENERIC,
+			     _("Image loading failed."));
+	}
+
 	return !failed;
 }
 
@@ -1062,6 +1218,13 @@ eog_image_has_data (EogImage *img, guint req_data)
 #endif
 	}
 
+	if ((req_data & EOG_IMAGE_DATA_XMP) > 0) {
+		req_data = (req_data & !EOG_IMAGE_DATA_XMP);
+#ifdef HAVE_EXEMPI
+		has_data = has_data && (priv->xmp != NULL);
+#endif
+	}
+
 	if (req_data != 0) {
 		g_warning ("Asking for unknown data, remaining: %i\n", req_data);
 		has_data = FALSE;
@@ -1379,6 +1542,11 @@ eog_image_reset_modifications (EogImage *image)
 		priv->trans = NULL;
 	}
 
+	if (priv->trans_autorotate != NULL) {
+		g_object_unref (priv->trans_autorotate);
+		priv->trans_autorotate = NULL;
+	}
+
 	priv->modified = FALSE;
 }
 
@@ -1429,6 +1597,17 @@ eog_image_save_by_info (EogImage *img, EogImageSaveInfo *source, GError **error)
 
 	priv = img->priv;
 
+	/* FIXME: Because of a bug in gnome-vfs file monitoring, we need to avoid
+         * an image that is being saved to be removed and re-added by the monitor
+         * handler. We do this by setting the image's status to "saving" and, in
+         * case the image is being monitored, the status is only restored after the
+         * "fake" image deletion is notified by gnome-vfs. Same aproach is used in
+         * eog_image_save_as_by_info(). See bug #46830. */
+	priv->prev_status = priv->status;
+
+	/* Image is now being saved */
+	priv->status = EOG_IMAGE_STATUS_SAVING;
+
 	/* see if we need any saving at all */
 	if (source->exists && !source->modified) {
 		return TRUE;
@@ -1477,6 +1656,11 @@ eog_image_save_by_info (EogImage *img, EogImageSaveInfo *source, GError **error)
 
 	g_free (tmpfile);
 
+	/* Restore previous status only if is not monitored image */
+	if (!priv->is_monitored) {
+		priv->status = priv->prev_status;
+	}
+
 	return success;
 }
 
@@ -1562,6 +1746,8 @@ eog_image_save_as_by_info (EogImage *img, EogImageSaveInfo *source, EogImageSave
 
 	priv = img->priv;
 
+	priv->prev_status = priv->status;
+
 	/* fail if there is no image to save */
 	if (priv->image == NULL) {
 		g_set_error (error, 
@@ -1616,6 +1802,11 @@ eog_image_save_as_by_info (EogImage *img, EogImageSaveInfo *source, EogImageSave
 	tmp_file_delete (tmpfile);
 	g_free (tmpfile);
 
+	/* Restore previous status only if is not monitored image */
+	if (!priv->is_monitored) {
+		priv->status = priv->prev_status;
+	}
+
 	return success;
 }
 
@@ -1668,9 +1859,6 @@ eog_image_get_caption (EogImage *img)
 		/* Use cached caption string */
 		return priv->caption;
 	
-	/* g_print ("priv->uri: %p\n", priv->uri);
-	   g_print ("%s\n", gnome_vfs_uri_to_string (priv->uri, GNOME_VFS_URI_HIDE_NONE)); */
-
 	name = gnome_vfs_uri_extract_short_name (priv->uri);
 	
 	if (name != NULL && g_ascii_strcasecmp (priv->uri->method_string, "file") == 0) {
@@ -1783,6 +1971,27 @@ eog_image_get_exif_info (EogImage *img)
 	return data;
 }
 
+
+gpointer
+eog_image_get_xmp_info (EogImage *img)
+{
+	EogImagePrivate *priv;
+ 	gpointer data = NULL;
+ 	
+ 	g_return_val_if_fail (EOG_IS_IMAGE (img), NULL);
+ 	
+ 	priv = img->priv;
+	
+#ifdef HAVE_EXEMPI
+ 	g_mutex_lock (priv->status_mutex);
+ 	data = (gpointer) xmp_copy (priv->xmp);
+ 	g_mutex_unlock (priv->status_mutex);
+#endif
+	
+ 	return data;
+}
+
+
 GnomeVFSURI *
 eog_image_get_uri (EogImage *img)
 {
@@ -1838,6 +2047,30 @@ eog_image_get_uri_for_display (EogImage *img)
 	return str;
 }
 
+EogImageStatus
+eog_image_get_status (EogImage *img)
+{
+	g_return_val_if_fail (EOG_IS_IMAGE (img), EOG_IMAGE_STATUS_UNKNOWN);
+
+	return img->priv->status;
+}
+
+void
+eog_image_restore_status (EogImage *img)
+{
+	g_return_if_fail (EOG_IS_IMAGE (img));
+
+	img->priv->status = img->priv->prev_status;
+}
+
+void
+eog_image_set_is_monitored (EogImage *img, gboolean is_monitored)
+{
+	g_return_if_fail (EOG_IS_IMAGE (img));
+
+	img->priv->is_monitored = is_monitored;
+}
+
 void
 eog_image_data_ref (EogImage *img)
 {
